@@ -3,7 +3,7 @@
 ;; Copyright (C) 2019-2024 Free Software Foundation, Inc.
 
 ;; Author: Mattias Engdegård <mattiase@acm.org>
-;; Version: 2.0
+;; Version: 2.1
 ;; Package-Requires: ((emacs "27.1"))
 ;; URL: https://github.com/mattiase/xr
 ;; Keywords: lisp, regexps
@@ -78,6 +78,19 @@ END is nil if unknown."
        (setq list (cdr list)))
      (not list)))
 
+(eval-when-compile
+  (defconst xr--char-classes '( ascii alnum alpha blank cntrl digit graph
+                                lower multibyte nonascii print punct space
+                                unibyte upper word xdigit)))
+
+;; FIXME: `eval-and-compile' around `pcase-defmacro' only necessary
+;; for compatibility with Emacs 27.
+(eval-and-compile
+  (pcase-defmacro xr--char-class ()
+    "Match any standard regexp char class as a symbol."
+    `(or ,@(mapcar (lambda (x) `(quote ,x)) xr--char-classes)))
+  )
+
 (defvar xr--string)
 (defvar xr--len)
 (defvar xr--idx)
@@ -118,11 +131,7 @@ END is nil if unknown."
              (let ((i (xr--string-search ":]" string (+ 2 idx))))
                (and i
                     (let ((sym (intern (substring string (+ idx 2) i))))
-                      (unless
-                          (memq sym
-                                '( ascii alnum alpha blank cntrl digit graph
-                                   lower multibyte nonascii print punct space
-                                   unibyte upper word xdigit))
+                      (unless (memq sym (eval-when-compile xr--char-classes))
                         (xr--error idx (1+ i)
                                    "No character class `[:%s:]'"
                                    (symbol-name sym)))
@@ -341,6 +350,8 @@ END is nil if unknown."
                (null classes))
           'anything)
          ;; Non-negated single-char set, like [$]: make a string.
+         ;; FIXME: Translate [^a] to (not "a") instead of (not (any "a")),
+         ;; the latter only required for use in Emacs 26 and older.
          ((and (= (length chars) 1)
                (not negated)
                (null ranges)
@@ -385,7 +396,8 @@ adjacent strings. SEQUENCE is used destructively."
             (rest (cdr sequence)))
         (setq sequence
               (cond ((stringp elem)
-                     (push elem strings)
+                     (unless (equal elem "")
+                       (push elem strings))
                      rest)
                     ((eq (car-safe elem) 'seq)
                      (nconc (nreverse (cdr elem)) rest))
@@ -1162,10 +1174,7 @@ nil if RX only matches the empty string."
          (and (xr--some #'xr--matches-nonempty body) 'sometimes)
        (xr--tristate-some #'xr--matches-nonempty body)))
     (`(,(or 'any 'not 'intersection 'syntax 'category) . ,_) 'always)
-    ((or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
-         'lower 'multibyte 'nonascii 'print 'punct 'space
-         'unibyte 'upper 'word 'xdigit
-         'nonl 'anything)
+    ((or 'nonl 'anything (xr--char-class))
      'always)))
 
 (defun xr--starts-with-sym (symbol item)
@@ -1490,14 +1499,9 @@ A-SETS and B-SETS are arguments to `any'."
     (`(not (any . ,b-sets))
      (and negated
           (xr--char-superset-of-char-set-p b-sets nil sets)))
-    ((or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
-         'lower 'multibyte 'nonascii 'print 'punct 'space
-         'unibyte 'upper 'word 'xdigit)
+    ((xr--char-class)
      (xr--char-superset-of-char-set-p sets negated (list rx)))
-    (`(not ,(and sym
-                 (or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
-                     'lower 'multibyte 'nonascii 'print 'punct 'space
-                     'unibyte 'upper 'word 'xdigit)))
+    (`(not ,(and sym (xr--char-class)))
      (and negated
           (xr--char-superset-of-char-set-p (list sym) nil sets)))
     ((pred characterp)
@@ -1513,11 +1517,9 @@ A-SETS and B-SETS are arguments to `any'."
 
 (defun xr--single-char-p (rx)
   "Whether RX only matches single characters."
-  (or (memq rx '(nonl anything
-                 ascii alnum alpha blank cntrl digit graph
-                 lower multibyte nonascii print punct space
-                 unibyte upper word xdigit
-                 wordchar not-wordchar))
+  (or (memq rx (eval-when-compile
+                 (append '(nonl anything wordchar not-wordchar)
+                         xr--char-classes)))
       (characterp rx)
       (and (consp rx)
            (or (memq (car rx) '(any category syntax))
@@ -1605,10 +1607,7 @@ A-SETS and B-SETS are arguments to `any'."
                        (and (characterp b)
                             (string-match-p (rx-to-string a)
                                             (char-to-string b)))))
-                  ((memq a-not-arg
-                         '( ascii alnum alpha blank cntrl digit graph
-                            lower multibyte nonascii print punct space
-                            unibyte upper word xdigit))
+                  ((memq a-not-arg (eval-when-compile xr--char-classes))
                    (xr--char-superset-of-rx-p (list a-not-arg) t b))
                   (t (equal a b)))))
 
@@ -1626,26 +1625,29 @@ A-SETS and B-SETS are arguments to `any'."
          ((eq a-op 'or)
           (xr--some (lambda (a-expr) (xr--superset-p a-expr b)) a-body))
 
-         ((eq a-op 'zero-or-more)
-          (if (memq (car-safe b) '(opt zero-or-more one-or-more))
-              (let ((b-body (cdr b)))
-                (xr--superset-p (xr--make-seq a-body) (xr--make-seq b-body)))
-            (xr--superset-p (xr--make-seq a-body) b)))
-         ((eq a-op 'one-or-more)
-          (if (eq (car-safe b) 'one-or-more)
-              (let ((b-body (cdr b)))
-                (xr--superset-p (xr--make-seq a-body) (xr--make-seq b-body)))
-            (xr--superset-p (xr--make-seq a-body) b)))
-         ((eq a-op 'opt)
-          (if (eq (car-safe b) 'opt)
+         ((memq a-op '(zero-or-more *?))
+          (if (memq (car-safe b) '(opt zero-or-more one-or-more ?? *? +?))
+            (let ((b-body (cdr b)))
+              (xr--superset-p (xr--make-seq a-body) (xr--make-seq b-body)))
+            (or (equal b '(seq))
+                (xr--superset-p (xr--make-seq a-body) b))))
+         ((memq a-op '(one-or-more +?))
+          (if (memq (car-safe b) '(one-or-more +?))
               (let ((b-body (cdr b)))
                 (xr--superset-p (xr--make-seq a-body) (xr--make-seq b-body)))
             (xr--superset-p (xr--make-seq a-body) b)))
+         ((memq a-op '(opt ??))
+          (if (memq (car-safe b) '(opt ??))
+            (let ((b-body (cdr b)))
+              (xr--superset-p (xr--make-seq a-body) (xr--make-seq b-body)))
+            (or (equal b '(seq))
+                (xr--superset-p (xr--make-seq a-body) b))))
          ((eq a-op 'repeat)
           (let ((lo (car a-body))
                 (a-body (cddr a-body)))
             (if (<= lo 1)
-                (xr--superset-p (xr--make-seq a-body) b)
+                (or (and (= lo 0) (equal b '(seq)))
+                    (xr--superset-p (xr--make-seq a-body) b))
               (equal a b))))
          
          ;; We do not expand through groups on the subset (b) side to
@@ -1661,9 +1663,7 @@ A-SETS and B-SETS are arguments to `any'."
 
          (t (equal a b)))))
 
-     ((memq a '( ascii alnum alpha blank cntrl digit graph
-                 lower multibyte nonascii print punct space
-                 unibyte upper word xdigit))
+     ((memq a (eval-when-compile xr--char-classes))
       (xr--char-superset-of-rx-p (list a) nil b))
 
      ((eq a 'nonl) (xr--single-non-newline-char-p b))
@@ -1685,10 +1685,7 @@ A-SETS and B-SETS are arguments to `any'."
   ;; `or'-expressions to make an `any' form.
   (pcase x
     ((pred stringp) (= (length x) 1))
-    ((or 'ascii 'alnum 'alpha 'blank 'cntrl 'digit 'graph
-         'lower 'multibyte 'nonascii 'print 'punct 'space
-         'unibyte 'upper 'word 'xdigit
-         'anything)
+    ((or (xr--char-class) 'anything)
      t)
     (`(any . ,_) t)
     ;; Assume for this purpose that \sw and \s- are equivalent to
@@ -1717,26 +1714,31 @@ A-SETS and B-SETS are arguments to `any'."
                      (let ((branch (car alts)))
                        (cond
                         ((xr--superset-p seq branch)
-                         (let ((duplicate (equal seq branch)))
+                         (let ((duplicate (equal seq branch))
+                               (prev-beg (cadr locs))
+                               (prev-end (- (car locs) 3)))
                            (xr--warn
                             warnings
-                            pos (1- xr--idx)
+                            pos (and (< pos xr--idx) (1- xr--idx))
                             (if duplicate
                                 "Duplicated alternative branch"
                               "Branch matches superset of a previous branch")
-                            (cadr locs) (- (car locs) 3)
+                            prev-beg (and (>= prev-end prev-beg) prev-end)
                             (if duplicate
                                 "Previous occurrence here"
                               "This is the subset branch"))
                            nil))
                         ((xr--superset-p branch seq)
-                         (xr--warn warnings
-                                   pos (1- xr--idx)
-                                   "Branch matches subset of a previous branch"
-                                   (cadr locs) (- (car locs) 3)
-                                   "This is the superset branch")
-                         nil)
-                        (t t))))
+                         (let ((prev-beg (cadr locs))
+                               (prev-end (- (car locs) 3)))
+                           (xr--warn
+                            warnings
+                            pos (and (< pos xr--idx) (1- xr--idx))
+                            "Branch matches subset of a previous branch"
+                            prev-beg (and (>= prev-end prev-beg) prev-end)
+                            "This is the superset branch")
+                           nil))
+                         (t t))))
               (setq locs (cdr locs))
               (setq alts (cdr alts))))
           (when (and (eq checks 'all)
@@ -1822,11 +1824,7 @@ A-SETS and B-SETS are arguments to `any'."
            (let ((i (xr--string-search ":]" string (+ 2 idx))))
              (and i
                   (let ((sym (intern (substring string (+ idx 2) i))))
-                    (unless
-                        (memq sym
-                              '( ascii alnum alpha blank cntrl digit graph
-                                 lower multibyte nonascii print punct space
-                                 unibyte upper word xdigit))
+                    (unless (memq sym (eval-when-compile xr--char-classes))
                       (xr--error idx (1+ i)
                                  "No character class `[:%s:]'"
                                  (symbol-name sym)))
